import colorama


def colorize(text, color):
    color_code = getattr(colorama.Fore, color.upper())
    return '%s%s%s' % (color_code, text, colorama.Fore.RESET)


def red(text):
    return colorize(text, 'red')


def green(text):
    return colorize(text, 'green')


def rangify(number_list):
    """Assumes the list is sorted."""
    if not number_list:
        return number_list

    ranges = []

    range_start = prev_num = number_list[0]
    for num in number_list[1:]:
        if num != (prev_num + 1):
            ranges.append((range_start, prev_num))
            range_start = num
        prev_num = num

    ranges.append((range_start, prev_num))
    return ranges



import os

import lxml.etree as ET


class Cobertura(object):
    """
    An XML cobertura parser.
    """

    def __init__(self, xml_source):
        """
        Initialize a cobertura report given an XML file `xml_source` that is
        in the cobertura format.

        `xml_source` may be:
        - a path to an XML file
        - an open file object
        - an XML string
        """
        self.source = xml_source

        self._classes = None
        self._packages = None

        self._lines_missed = {}
        self._lines_hit = {}

        self._package_list = None
        self._class_list = None

        self.xml = self._parse(xml_source)

    def _parse(self, xml_source):
        if hasattr(xml_source, 'startswith'):
            if xml_source.startswith('<?xml'):
                return self._parse_xml_string(xml_source)

            elif os.path.exists(xml_source):
                return self._parse_xml_file(xml_source)

        elif hasattr(xml_source, 'read'):
            return self._parse_xml_fileobj(xml_source)

    def _parse_xml_file(self, xml_file):
        return ET.parse(xml_file).getroot()

    def _parse_xml_fileobj(self, xml_fileobj):
        return self._parse_xml_string(xml_fileobj.read())

    def _parse_xml_string(self, xml_string):
        return ET.fromstring(xml_string)

    def _get_element_by_class_name(self, class_name):
        return self.xml.xpath("//class[@name='%s'][1]" % class_name)[0]

    def _get_lines_by_class_name(self, class_name):
        el = self._get_element_by_class_name(class_name)
        return el.xpath('./lines/line')

    @property
    def version(self):
        """Return the version number of the coverage report."""
        return self.xml.attrib['version']

    def line_rate(self, class_name=None):
        """
        Return the global line rate of the coverage report. If the class
        `class_name` is given, return the line rate of the class.
        """
        if class_name is None:
            el = self.xml
        else:
            el = self._get_element_by_class_name(class_name)

        return float(el.attrib['line-rate'])

    def branch_rate(self, class_name=None):
        """
        Return the global branch rate of the coverage report. If the class
        `class_name` is given, return the branch rate of the class.
        """
        if class_name is None:
            el = self.xml
        else:
            el = self._get_element_by_class_name(class_name)

        return float(el.attrib['branch-rate'])

    def missed_lines(self, class_name):
        """
        Return the list of uncovered line numbers for the class `class_name`.
        """
        line_elements = self._get_lines_by_class_name(class_name)
        lines = []
        prev_was_missed = False
        for line in line_elements:
            if line.attrib['hits'] != '0':
                prev_was_missed = False
                continue

            lineno = int(line.attrib['number'])
            if prev_was_missed is False:
                lines.append(lineno)
                prev_was_missed = True
            else:
                # A missed line (hits=0) in the Cobertura report actually
                # indicates a missed statement. A multiline statement will only
                # appear as 1 uncovered line in the source code (the line on
                # which the statement started) so we need to backfill the rest
                # of the lines part of the same statement by showing them as
                # missing too.
                missed_range = range(lines[-1]+1, lineno+1)
                lines += missed_range
        return lines

    def line_hits(self, class_name):
        """
        Return the list of covered line numbers for the class `class_name`.
        """
        el = self._get_element_by_class_name(class_name)
        lines = el.xpath('./lines/line[@hits>0]')
        return [int(l.attrib['number']) for l in lines]

    def total_misses(self, class_name):
        """
        Return the number of uncovered line numbers for the class `class_name`.
        """
        el = self._get_element_by_class_name(class_name)
        statements = el.xpath('./lines/line[@hits=0]')
        return len(statements)

    def total_hits(self, class_name):
        """
        Return the number of covered line numbers for the class `class_name`.
        """
        return len(self.line_hits(class_name))

    def total_lines(self, class_name):
        """
        Return the total number of lines for the class `class_name`.
        """
        el = self._get_element_by_class_name(class_name)
        lines = el.xpath('./lines/line')
        return len(lines)

    def filename(self, class_name):
        """
        Return the filename of the class `class_name`.
        """
        el = self._get_element_by_class_name(class_name)
        filename = el.attrib['filename']
        return filename

    def classes(self):
        """
        Return the list of available classes in the coverage report.
        """
        return [el.attrib['name'] for el in self.xml.xpath("//class")]

    def has_class(self, class_name):
        """
        Return `True` if the class `class_name` is present in the report,
        return `False` otherwise.
        """
        # FIXME: this will lookup a list which is slow, make it O(1)
        return class_name in self.classes()

    def packages(self):
        """
        Return the list of available packages in the coverage report.
        """
        return [el.attrib['name'] for el in self.xml.xpath("//package")]



import mock


SOURCE_FILE = 'tests/cobertura.xml'


def remove_style_tag(html):
    style_pattern_start = '\n    <style>'
    style_pattern_stop = '\n    </style>'
    style_starts = html.find(style_pattern_start)
    style_stops = html.find(style_pattern_stop) + len(style_pattern_stop)
    html_nostyle = html[:style_starts] + html[style_stops:]
    return html_nostyle


def make_cobertura(xml_file=SOURCE_FILE):
    from pycobertura import Cobertura
    cobertura = Cobertura(xml_file)
    return cobertura


def test_report_row_by_class():
    from pycobertura.reporters import TextReporter

    cobertura = make_cobertura()
    report = TextReporter(cobertura)

    expected_rows = {
        'Main': ['Main', 11, 0, 1, []],
        'search.BinarySearch': ['search.BinarySearch', 12, 1, 0.9166666666666666, [24]],
        'search.ISortedArraySearch': ['search.ISortedArraySearch', 0, 0, 1.0, []],
        'search.LinearSearch': ['search.LinearSearch', 7, 2, 0.7142857142857143, [19, 20, 21, 22, 23, 24]],
    }

    for class_name in cobertura.classes():
        row = report.get_report_row_by_class(class_name)
        assert row == expected_rows[class_name]


def test_text_report():
    from pycobertura.reporters import TextReporter

    cobertura = make_cobertura()
    report = TextReporter(cobertura)

    assert report.generate() == """\
Name                         Stmts    Miss  Cover    Missing
-------------------------  -------  ------  -------  ---------
Main                            11       0  100.00%
search.BinarySearch             12       1  91.67%   24
search.ISortedArraySearch        0       0  100.00%
search.LinearSearch              7       2  71.43%   19-24
TOTAL                           30       3  90.00%"""


def test_text_report__with_missing_range():
    from pycobertura.reporters import TextReporter

    cobertura = make_cobertura('tests/dummy.with-dummy2-no-cov.xml')
    report = TextReporter(cobertura)

    assert report.generate() == """\
Name              Stmts    Miss  Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__        0       0  0.00%
dummy/dummy           4       0  100.00%
dummy/dummy2          2       2  0.00%    1-2
TOTAL                 6       2  66.67%"""


def test_text_report_delta__no_diff():
    from pycobertura.reporters import TextReporter, TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.original.xml')
    cobertura2 = make_cobertura('tests/dummy.original.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2)

    assert report_delta.generate() == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        -       -
TOTAL           -        -       -"""


def test_text_report_delta__improve_coverage():
    from pycobertura.reporters import TextReporter, TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.original.xml')
    cobertura2 = make_cobertura('tests/dummy.original-better-cov.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2)

    assert report_delta.generate() == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        -1      +25.00%  -2
TOTAL           -        -1      +25.00%"""


def test_text_report_delta__full_coverage():
    from pycobertura.reporters import TextReporter, TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.original.xml')
    cobertura2 = make_cobertura('tests/dummy.original-full-cov.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2)

    assert report_delta.generate() == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        -2      +50.00%  -2, -5
TOTAL           -        -2      +50.00%"""


def test_text_report_delta__worsen_coverage():
    from pycobertura.reporters import TextReporter, TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.original-better-cov.xml')
    cobertura2 = make_cobertura('tests/dummy.original.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2)

    assert report_delta.generate() == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        +1      -25.00%  +2
TOTAL           -        +1      -25.00%"""


def test_text_report_delta__new_dummy2_class_has_no_coverage():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.original-full-cov.xml')
    cobertura2 = make_cobertura('tests/dummy.with-dummy2-no-cov.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2)

    assert report_delta.generate() == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        -       -
dummy/dummy2    +2       +2      -        +1, +2
TOTAL           +2       +2      -33.33%"""


def test_text_report_delta__new_dummy2_class_has_improved_coverage():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.original-full-cov.xml')
    cobertura2 = make_cobertura('tests/dummy.with-dummy2-better-cov.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2)

    assert report_delta.generate() == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        -       -
dummy/dummy2    +2       +1      +50.00%  +2
TOTAL           +2       +1      -16.67%"""


def test_text_report_delta__new_dummy2_class_has_full_coverage():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.original-full-cov.xml')
    cobertura2 = make_cobertura('tests/dummy.with-dummy2-full-cov.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2)

    assert report_delta.generate() == """\
Name            Stmts    Miss    Cover     Missing
--------------  -------  ------  --------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        -       -
dummy/dummy2    +2       -       +100.00%
TOTAL           +2       -       -"""


def test_text_report_delta__removed_dummy2_class():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.with-dummy2-full-cov.xml')
    cobertura2 = make_cobertura('tests/dummy.original-full-cov.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2)

    assert report_delta.generate() == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        -       -
dummy/dummy2    -2       -       -
TOTAL           -2       -       -"""


def test_text_report_delta__removed_dummy2_class_and_worsen_coverage():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.with-dummy2-full-cov.xml')
    cobertura2 = make_cobertura('tests/dummy.original-better-cov.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2)

    assert report_delta.generate() == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        +1      -25.00%  +5
dummy/dummy2    -2       -       -
TOTAL           -2       +1      -25.00%"""


def test_text_report_delta__better_and_worse_coverage():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.with-dummy2-better-cov.xml')
    cobertura2 = make_cobertura('tests/dummy.with-dummy2-better-and-worse.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2)

    assert report_delta.generate() == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        +1      -25.00%  +5
dummy/dummy2    -        -1      +50.00%  -2
TOTAL           -        -       -"""


def test_text_report_delta__colorize_True():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.with-dummy2-better-cov.xml')
    cobertura2 = make_cobertura('tests/dummy.with-dummy2-better-and-worse.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2, color=True)

    assert report_delta.generate() == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        +1      -25.00%  \x1b[31m+5\x1b[39m
dummy/dummy2    -        -1      +50.00%  \x1b[32m-2\x1b[39m
TOTAL           -        -       -"""


def test_text_report_delta__colorize_True__with_missing_range():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.original.xml')
    cobertura2 = make_cobertura('tests/dummy.with-dummy2-no-cov.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2, color=True)

    assert report_delta.generate() == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        -2      +50.00%  \x1b[32m-2\x1b[39m, \x1b[32m-5\x1b[39m
dummy/dummy2    +2       +2      -        \x1b[31m+1\x1b[39m, \x1b[31m+2\x1b[39m
TOTAL           +2       -       +16.67%"""


def test_text_report_delta__colorize_False():
    from pycobertura.reporters import TextReporterDelta

    cobertura1 = make_cobertura('tests/dummy.original.xml')
    cobertura2 = make_cobertura('tests/dummy.with-dummy2-no-cov.xml')

    report_delta = TextReporterDelta(cobertura1, cobertura2, color=False)

    assert report_delta.generate() == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        -2      +50.00%  -2, -5
dummy/dummy2    +2       +2      -        +1, +2
TOTAL           +2       -       +16.67%"""


def test_html_report():
    from pycobertura.reporters import HtmlReporter

    cobertura = make_cobertura()
    report = HtmlReporter(cobertura)
    html_output = report.generate()

    assert "normalize.css" in html_output
    assert "Skeleton V2.0" in html_output

    assert remove_style_tag(html_output) == """\
<html>
  <head>
    <title>pycobertura report</title>
    <meta charset="UTF-8">
  </head>
  <body>
    <table>
      <thead>
        <tr>
          <th>Name</th>
          <th>Stmts</th>
          <th>Miss</th>
          <th>Cover</th>
          <th>Missing</th>
        </tr>
      </thead>
      <tbody>
        <tr>
          <td>Main</td>
          <td>11</td>
          <td>0</td>
          <td>100.00%</td>
          <td></td>
        </tr>
        <tr>
          <td>search.BinarySearch</td>
          <td>12</td>
          <td>1</td>
          <td>91.67%</td>
          <td>24</td>
        </tr>
        <tr>
          <td>search.ISortedArraySearch</td>
          <td>0</td>
          <td>0</td>
          <td>100.00%</td>
          <td></td>
        </tr>
        <tr>
          <td>search.LinearSearch</td>
          <td>7</td>
          <td>2</td>
          <td>71.43%</td>
          <td>19-24</td>
        </tr>
      </tbody>
      <tfoot>
        <tr>
          <td>TOTAL</td>
          <td>30</td>
          <td>3</td>
          <td>90.00%</td>
          <td></td>
        </tr>
      </tfoot>
    </table>
  </body>
</html>"""


def test_html_report_delta():
    from pycobertura.reporters import HtmlReporterDelta

    cobertura1 = make_cobertura('tests/dummy.original-full-cov.xml')
    cobertura2 = make_cobertura('tests/dummy.with-dummy2-no-cov.xml')

    report_delta = HtmlReporterDelta(cobertura1, cobertura2)
    html_output = report_delta.generate()
    assert '.red {color: red}' in html_output
    assert '.green {color: green}' in html_output
    assert "normalize.css" in html_output
    assert "Skeleton V2.0" in html_output

    assert remove_style_tag(html_output) == """\
<html>
  <head>
    <title>pycobertura report</title>
    <meta charset="UTF-8">
  </head>
  <body>
    <table>
      <thead>
        <tr>
          <th>Name</th>
          <th>Stmts</th>
          <th>Miss</th>
          <th>Cover</th>
          <th>Missing</th>
        </tr>
      </thead>
      <tbody>
        <tr>
          <td>dummy/__init__</td>
          <td>-</td>
          <td>-</td>
          <td>-</td>
          <td>
          </td>
        </tr>
        <tr>
          <td>dummy/dummy</td>
          <td>-</td>
          <td>-</td>
          <td>-</td>
          <td>
          </td>
        </tr>
        <tr>
          <td>dummy/dummy2</td>
          <td>+2</td>
          <td>+2</td>
          <td>-</td>
          <td><span class="red">+1</span>, <span class="red">+2</span>
          </td>
        </tr>
      </tbody>
      <tfoot>
        <tr>
          <td>TOTAL</td>
          <td>+2</td>
          <td>+2</td>
          <td>-33.33%</td>
          <td></td>
        </tr>
      </tfoot>
    </table>
  </body>
</html>"""



import mock

import lxml.etree as ET


SOURCE_FILE = 'tests/cobertura.xml'


def make_cobertura():
    from pycobertura import Cobertura
    cobertura = Cobertura(SOURCE_FILE)
    return cobertura


def test_parse_string():
    from pycobertura import Cobertura

    with open(SOURCE_FILE) as f:
        xml = f.read()

    cobertura = Cobertura(xml)
    assert isinstance(cobertura.xml, ET._Element)


def test_parse_fileobj():
    from pycobertura import Cobertura

    with open(SOURCE_FILE) as f:
        cobertura = Cobertura(f)
    assert isinstance(cobertura.xml, ET._Element)


def test_parse_path():
    from pycobertura import Cobertura

    xml_path = 'foo.xml'

    with mock.patch('pycobertura.cobertura.os.path.exists', return_value=True):
        with mock.patch('pycobertura.cobertura.ET.parse') as mock_parse:
            cobertura = Cobertura(xml_path)

    assert cobertura.xml is mock_parse.return_value.getroot.return_value


def test_version():
    cobertura = make_cobertura()
    assert cobertura.version == '1.9'


def test_line_rate():
    cobertura = make_cobertura()
    assert cobertura.line_rate() == 0.9


def test_line_rate_by_class():
    cobertura = make_cobertura()
    expected_line_rates = {
        'Main': 1.0,
        'search.BinarySearch': 0.9166666666666666,
        'search.ISortedArraySearch': 1.0,
        'search.LinearSearch': 0.7142857142857143,
    }

    for class_name in cobertura.classes():
        assert cobertura.line_rate(class_name) == \
            expected_line_rates[class_name]


def test_branch_rate():
    cobertura = make_cobertura()
    assert cobertura.branch_rate() == 0.75


def test_branch_rate_by_class():
    cobertura = make_cobertura()
    expected_branch_rates = {
        'Main': 1.0,
        'search.BinarySearch': 0.8333333333333334,
        'search.ISortedArraySearch': 1.0,
        'search.LinearSearch': 0.6666666666666666,
    }

    for class_name in cobertura.classes():
        assert cobertura.branch_rate(class_name) == \
            expected_branch_rates[class_name]


def test_list_packages():
    cobertura = make_cobertura()

    packages = cobertura.packages()
    assert packages == ['', 'search']


def test_list_classes():
    cobertura = make_cobertura()

    classes = cobertura.classes()
    assert classes == [
        'Main',
        'search.BinarySearch',
        'search.ISortedArraySearch',
        'search.LinearSearch'
    ]


def test_line_hits__by_iterating_over_classes():
    cobertura = make_cobertura()

    expected_lines = {
        'Main': [10, 16, 17, 18, 19, 23, 25, 26, 28, 29, 30],
        'search.BinarySearch': [12, 16, 18, 20, 21, 23, 25, 26, 28, 29, 31],
        'search.ISortedArraySearch': [],
        'search.LinearSearch': [9, 13, 15, 16, 17],
    }

    for class_name in cobertura.classes():
        assert cobertura.line_hits(class_name) == expected_lines[class_name]


def test_line_misses():
    cobertura = make_cobertura()

    expected_lines = {
        'Main': [],
        'search.BinarySearch': [24],
        'search.ISortedArraySearch': [],
        'search.LinearSearch': [19, 20, 21, 22, 23, 24],
    }

    for class_name in cobertura.classes():
        assert cobertura.missed_lines(class_name) == expected_lines[class_name]


def test_total_lines():
    cobertura = make_cobertura()
    expected_total_lines = {
        'Main': 11,
        'search.BinarySearch': 12,
        'search.ISortedArraySearch': 0,
        'search.LinearSearch': 7,
    }
    for class_name in cobertura.classes():
        assert cobertura.total_lines(class_name) == \
            expected_total_lines[class_name]


def test_total_misses():
    cobertura = make_cobertura()
    expected_total_misses = {
        'Main': 0,
        'search.BinarySearch': 1,
        'search.ISortedArraySearch': 0,
        'search.LinearSearch': 2,
    }
    for class_name in cobertura.classes():
        assert cobertura.total_misses(class_name) == \
            expected_total_misses[class_name]


def test_total_hits():
    cobertura = make_cobertura()
    expected_total_misses = {
        'Main': 11,
        'search.BinarySearch': 11,
        'search.ISortedArraySearch': 0,
        'search.LinearSearch': 5,
    }
    for class_name in cobertura.classes():
        assert cobertura.total_hits(class_name) == \
            expected_total_misses[class_name]


def test_filename():
    cobertura = make_cobertura()
    expected_filenames = {
        'Main': 'Main.java',
        'search.BinarySearch': 'search/BinarySearch.java',
        'search.ISortedArraySearch': 'search/ISortedArraySearch.java',
        'search.LinearSearch': 'search/LinearSearch.java',
    }
    for class_name in cobertura.classes():
        assert cobertura.filename(class_name) == \
            expected_filenames[class_name]



def test_rangify_func__1():
    from pycobertura.utils import rangify

    assert rangify([1]) == [(1, 1)]


def test_rangify_func__1_2():
    from pycobertura.utils import rangify

    assert rangify([1, 2]) == [(1, 2)]


def test_rangify_func__1_2_3():
    from pycobertura.utils import rangify

    assert rangify([1, 2, 3]) == [(1, 3)]


def test_rangify_func__1_2_3_and_7():
    from pycobertura.utils import rangify

    assert rangify([1, 2, 3, 7]) == [(1, 3), (7, 7)]


def test_rangify_func__1_2_3_and_7_8():
    from pycobertura.utils import rangify

    assert rangify([1, 2, 3, 7, 8]) == [(1, 3), (7, 8)]


def test_rangify_func__1_2_3_and_7_8_9():
    from pycobertura.utils import rangify

    assert rangify([1, 2, 3, 7, 8, 9]) == [(1, 3), (7, 9)]


def test_rangify_func__1_and_7_8_9():
    from pycobertura.utils import rangify

    assert rangify([1, 7, 8, 9]) == [(1, 1), (7, 9)]


def test_rangify_func__1_2_and_7_8_9():
    from pycobertura.utils import rangify

    assert rangify([1, 2, 7, 8, 9]) == [(1, 2), (7, 9)]


def test_rangify_func__1_and_7():
    from pycobertura.utils import rangify

    assert rangify([1, 7]) == [(1, 1), (7, 7)]



from click.testing import CliRunner
import sys

PY2 = sys.version_info[0] == 2


def test_show__format_default():
    from pycobertura.cli import show

    runner = CliRunner()
    result = runner.invoke(show,
        ['tests/dummy.original.xml'],
        catch_exceptions=False
    )
    assert result.output == """\
Name              Stmts    Miss  Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__        0       0  0.00%
dummy/dummy           4       2  50.00%   2, 5
TOTAL                 4       2  50.00%
"""


def test_show__format_text():
    from pycobertura.cli import show

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(
            show,
            ['tests/dummy.original.xml', opt, 'text'],
            catch_exceptions=False
        )
        assert result.output == """\
Name              Stmts    Miss  Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__        0       0  0.00%
dummy/dummy           4       2  50.00%   2, 5
TOTAL                 4       2  50.00%
"""


def test_show__format_html():
    from pycobertura.cli import show

    runner = CliRunner()
    result = runner.invoke(show, [
        'tests/dummy.original.xml', '--format', 'html'
    ], catch_exceptions=False)
    assert result.output.startswith('<html>')
    assert result.output.endswith('</html>\n')


def test_show__output_to_file():
    from pycobertura.cli import show

    with open('tests/dummy.original.xml') as f:
        cobertura_content = f.read()

    runner = CliRunner()
    for opt in ('-o', '--output'):
        with runner.isolated_filesystem():
            with open('cobertura.xml', 'w') as f:
                f.write(cobertura_content)
            result = runner.invoke(show, [
                'cobertura.xml', opt, 'report.out'
            ], catch_exceptions=False)
            with open('report.out') as f:
                report = f.read()
            assert result.output == ""
            assert report == """\
Name              Stmts    Miss  Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__        0       0  0.00%
dummy/dummy           4       2  50.00%   2, 5
TOTAL                 4       2  50.00%"""


def test_diff__format_default():
    from pycobertura.cli import diff

    runner = CliRunner()
    result = runner.invoke(diff, [
        'tests/dummy.with-dummy2-better-cov.xml',
        'tests/dummy.with-dummy2-better-and-worse.xml',
    ], catch_exceptions=False)
    # FIXME: Fails in PY2, requires click>=4.x to pass
    if not PY2:
        assert result.output == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        +1      -25.00%  \x1b[31m+5\x1b[39m
dummy/dummy2    -        -1      +50.00%  \x1b[32m-2\x1b[39m
TOTAL           -        -       -
"""
    else:
        assert result.output == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        +1      -25.00%  +5
dummy/dummy2    -        -1      +50.00%  -2
TOTAL           -        -       -
"""


def test_diff__format_text():
    from pycobertura.cli import diff

    runner = CliRunner()
    for opt in ('-f', '--format'):
        result = runner.invoke(diff, [
            opt, 'text',
            'tests/dummy.with-dummy2-better-cov.xml',
            'tests/dummy.with-dummy2-better-and-worse.xml',
        ], catch_exceptions=False)
        # FIXME: Fails in PY2, requires click>=4.x to pass
        if not PY2:
            assert result.output == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        +1      -25.00%  \x1b[31m+5\x1b[39m
dummy/dummy2    -        -1      +50.00%  \x1b[32m-2\x1b[39m
TOTAL           -        -       -
"""
        else:
            assert result.output == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        +1      -25.00%  +5
dummy/dummy2    -        -1      +50.00%  -2
TOTAL           -        -       -
"""


def test_diff__output_to_file():
    from pycobertura.cli import diff

    with open('tests/dummy.with-dummy2-better-cov.xml') as f:
        cobertura1_content = f.read()

    with open('tests/dummy.with-dummy2-better-and-worse.xml') as f:
        cobertura2_content = f.read()

    runner = CliRunner()

    for opt in ('-o', '--output'):
        with runner.isolated_filesystem():

            with open('cobertura1.xml', 'w') as f:
                f.write(cobertura1_content)

            with open('cobertura2.xml', 'w') as f:
                f.write(cobertura2_content)

            result = runner.invoke(diff, [
                'cobertura1.xml',
                'cobertura2.xml',
                opt, 'report.out'
            ], catch_exceptions=False)
            with open('report.out') as f:
                report = f.read()
            assert result.output == ""
            assert report == """\
Name            Stmts    Miss    Cover    Missing
--------------  -------  ------  -------  ---------
dummy/__init__  -        -       -
dummy/dummy     -        +1      -25.00%  +5
dummy/dummy2    -        -1      +50.00%  -2
TOTAL           -        -       -"""


# FIXME: when Click 4 is available, uncomment this.
#def test_diff__format_text__with_color():
#    from pycobertura.cli import diff
#
#    runner = CliRunner()
#    result = runner.invoke(diff, [
#        '--color',
#        'tests/dummy.with-dummy2-better-cov.xml',
#        'tests/dummy.with-dummy2-better-and-worse.xml',
#    ], catch_exceptions=False)
#    assert result.output == """\
#Name          Stmts    Miss    Cover    Missing
#------------  -------  ------  -------  ---------
#dummy/dummy   -        +1      -25.00%  \x1b[31m+5\x1b[39m
#dummy/dummy2  -        -1      +50.00%  \x1b[32m-2\x1b[39m
#TOTAL         -        -       -
#"""


def test_diff__format_html():
    from pycobertura.cli import diff

    runner = CliRunner()
    result = runner.invoke(diff, [
        '--format', 'html',
        'tests/dummy.with-dummy2-better-cov.xml',
        'tests/dummy.with-dummy2-better-and-worse.xml',
    ], catch_exceptions=False)
    assert result.output.startswith('<html>')
    assert result.output.endswith('</html>\n')


def test_diff__same_coverage_has_exit_status_of_zero():
    from pycobertura.cli import diff

    runner = CliRunner()
    result = runner.invoke(diff, [
        'tests/dummy.original.xml',
        'tests/dummy.original.xml',
    ], catch_exceptions=False)
    assert result.exit_code == 0


def test_diff__better_coverage_has_exit_status_of_zero():
    from pycobertura.cli import diff

    runner = CliRunner()
    result = runner.invoke(diff, [
        'tests/dummy.original.xml',
        'tests/dummy.original-full-cov.xml',
    ], catch_exceptions=False)
    assert result.exit_code == 0


def test_diff__worse_coverage_has_exit_status_of_one():
    from pycobertura.cli import diff

    runner = CliRunner()
    result = runner.invoke(diff, [
        'tests/dummy.original-full-cov.xml',
        'tests/dummy.original.xml',
    ], catch_exceptions=False)
    assert result.exit_code == 1



import os
import sys
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

version = '0.3.0'

try:
    import pypandoc
    README = pypandoc.convert('README.md', 'rst')
    CHANGES = pypandoc.convert('CHANGES.md', 'rst')
except:
    README = read('README.md')
    CHANGES = read('CHANGES.md')

setup(
    name="pycobertura",
    version=version,
    author="Alex Conrad",
    author_email="alexandre@surveymonkey.com",
    maintainer="Alex Conrad",
    maintainer_email="alexandre@surveymonkey.com",
    description="A Cobertura coverage parser that can diff reports.",
    license="MIT License",
    keywords="cobertura coverage parser parse xml",
    url="https://github.com/SurveyMonkey/pycobertura",
    zip_safe=False,
    include_package_data=True,
    packages=find_packages(exclude=['tests']),
    long_description='%s\n\n%s' % (README, CHANGES),
    setup_requires=['setuptools_git'],
    install_requires=['click', 'colorama', 'jinja2', 'lxml', 'tabulate'],
    classifiers=[
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4"
    ],
    entry_points = {
        'console_scripts': [
            'pycobertura=pycobertura.cli:pycobertura'
        ],
    },
)



import click

from pycobertura.cobertura import Cobertura
from pycobertura.reporters import (
    HtmlReporter, TextReporter, HtmlReporterDelta, TextReporterDelta
)


pycobertura = click.Group()


reporters = {
    'html': HtmlReporter,
    'text': TextReporter,
}


@pycobertura.command()
@click.argument('cobertura_file')
@click.option(
    '-f', '--format',
    default='text',
    type=click.Choice(list(reporters))
)
@click.option(
    '-o', '--output', metavar='<file>',
    type=click.File('wb'),
    help='Write output to <file> instead of stdout.'
)
def show(cobertura_file, format, output):
    """show the coverage summary"""
    cobertura = Cobertura(cobertura_file)
    Reporter = reporters[format]
    reporter = Reporter(cobertura)
    report = reporter.generate()

    if not isinstance(report, bytes):
        report = report.encode('utf-8')

    isatty = True if output is None else output.isatty()
    click.echo(report, file=output, nl=isatty)


delta_reporters = {
    'text': TextReporterDelta,
    'html': HtmlReporterDelta,
}


@pycobertura.command()
@click.argument('cobertura_file1')
@click.argument('cobertura_file2')
@click.option(
    '--color/--no-color', default=None,
    help='Colorize the output. By default, pycobertura emits color codes only '
         'when standard output is connected to a terminal. This has no effect '
         'with the HTML output format.')
@click.option(
    '-f', '--format', default='text',
    type=click.Choice(list(delta_reporters))
)
@click.option(
    '-o', '--output', metavar='<file>',
    type=click.File('wb'),
    help='Write output to <file> instead of stdout.'
)
def diff(cobertura_file1, cobertura_file2, color, format, output):
    """compare two coverage files"""
    cobertura1 = Cobertura(cobertura_file1)
    cobertura2 = Cobertura(cobertura_file2)
    Reporter = delta_reporters[format]
    reporter_args = [cobertura1, cobertura2]
    reporter_kwargs = {}

    isatty = True if output is None else output.isatty()
    if format == 'text':
        color = isatty if color is None else color is True
        reporter_kwargs['color'] = color

    reporter = Reporter(*reporter_args, **reporter_kwargs)
    report = reporter.generate()

    if not isinstance(report, bytes):
        report = report.encode('utf-8')

    # FIXME: In Click 3.x, the color is stripped out if the output file is
    # detected to not support ANSI codes causing `-o outfile --color` to write
    # to `outfile` without colors. In Click 4.x, `echo()` will take a
    # color=True/False flag which will allow us to override ANSI code
    # auto-detection.
    # https://github.com/mitsuhiko/click/commit/5cf7f2ddfba328070751cbda32782520d4e0d6f5
    # click.echo(report, file=output, nl=isatty, color=color)  # click 4.x
    click.echo(report, file=output, nl=isatty)

    # non-zero exit code if line rate worsened
    exit_code = cobertura1.line_rate() > cobertura2.line_rate()
    raise SystemExit(exit_code)



from jinja2 import Environment, PackageLoader
from pycobertura.utils import green, rangify, red
from tabulate import tabulate


env = Environment(loader=PackageLoader('pycobertura', 'templates'))


def get_class_summary_row(cobertura, class_name):
    if not cobertura.has_class(class_name):
        return None

    row = [
        class_name,
        cobertura.total_lines(class_name),
        cobertura.total_misses(class_name),
        cobertura.line_rate(class_name),
        cobertura.missed_lines(class_name),
    ]
    return row


class Reporter(object):
    def __init__(self, cobertura):
        self.cobertura = cobertura

    def get_report_row_by_class(self, class_name):
        total_lines = self.cobertura.total_lines(class_name)
        line_rate = self.cobertura.line_rate(class_name)
        total_misses = self.cobertura.total_misses(class_name)
        missed_lines = self.cobertura.missed_lines(class_name)

        row = [
            class_name,
            total_lines,
            total_misses,
            line_rate,
            missed_lines,
        ]

        return row

    def get_footer_row(self, lines):
        total_statements = 0
        total_misses = 0
        line_rates = []

        for class_name, total_lines, misses, line_rate, missed_lines in lines:
            total_statements += total_lines
            total_misses += misses
            line_rates.append(line_rate)

        total_line_rate = self.cobertura.line_rate()

        footer = [
            'TOTAL',
            total_statements,
            total_misses,
            total_line_rate,
            [],  # dummy missed lines
        ]

        return footer

    def get_report_lines(self):
        lines = []

        for class_name in self.cobertura.classes():
            row = get_class_summary_row(self.cobertura, class_name)
            lines.append(row)

        footer = self.get_footer_row(lines)
        lines.append(footer)

        return lines


class TextReporter(Reporter):
    def format_row(self, row):
        class_name, total_lines, total_misses, line_rate, missed_lines = row

        line_rate = '%.2f%%' % (line_rate * 100)

        formatted_missed_lines = []
        for line_start, line_stop in rangify(missed_lines):
            if line_start == line_stop:
                formatted_missed_lines.append('%s' % line_start)
            else:
                line_range = '%s-%s' % (line_start, line_stop)
                formatted_missed_lines.append(line_range)
        formatted_missed_lines = ', '.join(formatted_missed_lines)

        row = [
            class_name,
            total_lines,
            total_misses,
            line_rate,
            formatted_missed_lines,
        ]

        return row

    def generate(self):
        lines = self.get_report_lines()

        formatted_lines = []
        for row in lines:
            formatted_row = self.format_row(row)
            formatted_lines.append(formatted_row)

        report = tabulate(
            formatted_lines,
            headers=['Name', 'Stmts', 'Miss', 'Cover', 'Missing']
        )

        return report


class HtmlReporter(TextReporter):

    def generate(self):
        lines = self.get_report_lines()

        formatted_lines = []
        for row in lines:
            formatted_row = self.format_row(row)
            formatted_lines.append(formatted_row)

        template = env.get_template('html.jinja2')
        return template.render(
            lines=formatted_lines[:-1],
            footer=formatted_lines[-1]
        )


class DeltaReporter(object):
    def __init__(self, cobertura1, cobertura2):
        self.cobertura1 = cobertura1
        self.cobertura2 = cobertura2

    def get_diff_line(self, line1, line2):
        if line1 is not None:
            (class_name1,
             total_lines1,
             total_misses1,
             line_rate1,
             missed_lines1) = line1
            class_name = class_name1
        if line2 is not None:
            (class_name2,
             total_lines2,
             total_misses2,
             line_rate2,
             missed_lines2) = line2
            class_name = class_name2

        if None not in (line1, line2):
            added_lines = set(missed_lines2).difference(missed_lines1)
            removed_lines = set(missed_lines1).difference(missed_lines2)
            all_lines = sorted(added_lines.union(removed_lines))
        elif line1 is None:
            added_lines = all_lines = missed_lines2
        elif line2 is None:
            added_lines = set()
            all_lines = missed_lines1

        # Early return if line2 is None, the class doesn't exist anymore in
        # the latest version.
        if line2 is None:
            return [
                class_name,
                -total_lines1,  # negate line number of previous coverage
                0,  # total misses not applicable
                0,  # line rate not applicable
                [],  # missing lines not applicable
            ]

        formatted_lines = []
        for line in all_lines:
            sign = '+' if line in added_lines else '-'
            formatted_line = (sign, line)
            formatted_lines.append(formatted_line)

        if None not in (line1, line2):
            diff_line = [
                class_name,
                total_lines2 - total_lines1,
                total_misses2 - total_misses1,
                line_rate2 - line_rate1,
                formatted_lines,
            ]
        elif line1 is None:
            diff_line = [
                class_name,
                total_lines2,
                total_misses2,
                line_rate2,
                formatted_lines,
            ]

        return diff_line

    def get_footer_row(self, lines):
        total_statements = 0
        total_misses = 0
        line_rates = []

        for class_name, total_lines, misses, line_rate, missed_lines in lines:
            total_statements += total_lines
            total_misses += misses
            line_rates.append(line_rate)

        total_line_rate1 = self.cobertura1.line_rate()
        total_line_rate2 = self.cobertura2.line_rate()
        total_line_rate = total_line_rate2 - total_line_rate1

        footer = [
            'TOTAL',
            total_statements,
            total_misses,
            total_line_rate,
            [],  # dummy missed lines
        ]

        return footer

    def get_report_lines(self):
        lines = []

        classes1 = self.cobertura1.classes()
        classes2 = self.cobertura2.classes()
        all_classes = set(classes1).union(set(classes2))
        for class_name in sorted(all_classes):
            row1 = get_class_summary_row(self.cobertura1, class_name)
            row2 = get_class_summary_row(self.cobertura2, class_name)
            row = self.get_diff_line(row1, row2)
            lines.append(row)

        footer = self.get_footer_row(lines)
        lines.append(footer)

        return lines


class TextReporterDelta(DeltaReporter):
    def __init__(self, *args, **kwargs):
        """
        Takes the same arguments as `DeltaReporter` but also takes the keyword
        argument `color` which can be set to True or False depending if the
        generated report should be colored or not (default `color=False`).
        """
        self.color = kwargs.pop('color', False)
        super(TextReporterDelta, self).__init__(*args, **kwargs)

    def format_row(self, row):
        class_name, total_lines, total_misses, line_rate, missed_lines = row

        total_lines = '%+d' % total_lines if total_lines else '-'
        total_misses = '%+d' % total_misses if total_misses else '-'
        line_rate = '%+.2f%%' % (line_rate * 100) if line_rate else '-'
        missed_lines = ['%s%d' % l for l in missed_lines]

        if self.color is True:
            missed_lines_colored = []
            for line in missed_lines:
                colorize = [green, red][line[0] == '+']
                colored_line = colorize(line)
                missed_lines_colored.append(colored_line)
        else:
            missed_lines_colored = missed_lines

        missed_lines = ', '.join(missed_lines_colored)

        row = [
            class_name,
            total_lines,
            total_misses,
            line_rate,
            missed_lines,
        ]

        return row

    def generate(self):
        lines = self.get_report_lines()

        formatted_lines = []
        for row in lines:
            formatted_row = self.format_row(row)
            formatted_lines.append(formatted_row)

        report = tabulate(
            formatted_lines,
            headers=['Name', 'Stmts', 'Miss', 'Cover', 'Missing']
        )

        return report


class HtmlReporterDelta(TextReporterDelta):
    def format_row(self, row):
        class_name, total_lines, total_misses, line_rate, missed_lines = row

        total_lines = '%+d' % total_lines if total_lines else '-'
        total_misses = '%+d' % total_misses if total_misses else '-'
        line_rate = '%+.2f%%' % (line_rate * 100) if line_rate else '-'
        missed_lines = ['%s%d' % l for l in missed_lines]

        missed_lines_colored = []
        for line in missed_lines:
            colorize = [green, red][line[0] == '+']
            colored_line = colorize(line)
            missed_lines_colored.append(colored_line)

        row = [
            class_name,
            total_lines,
            total_misses,
            line_rate,
            missed_lines,
        ]

        return row

    def generate(self):
        lines = self.get_report_lines()

        formatted_lines = []
        for row in lines:
            formatted_row = self.format_row(row)
            formatted_lines.append(formatted_row)

        template = env.get_template('html-delta.jinja2')
        return template.render(
            lines=formatted_lines[:-1],
            footer=formatted_lines[-1]
        )
